/*
 * Example Code for the Paper
 *
 * "Mesh Saliency via Local Curvature Entropy", Eurographics 2016
 *
 * by M. Limper, A. Kuijper and D. Fellner
 *
 */

#include "lce.h"

#include <float.h>
#include <set>
#include <map>
#include <fstream>
#include <ctime>

#include "curvature.h"
#include "Histogram.h"
#include "FloatHistogram.h"
#include "utils.h"


#define NUM_SALIENCY_LEVELS 5
#define USE_VERTEX_AREA_WEIGHTING

//------------------------------------------------------------------------------------------------------------------

float computeSaliency(const lce::TriMesh & mesh, lce::TriMesh::VertexHandle v,
                      const std::vector<lce::TriMesh::VertexHandle> & neighbors,
                      OpenMesh::VPropHandleT<float> vertexAreasProperty,
                      OpenMesh::VPropHandleT<float> curvatureMeasure, float min, float max)
{
#ifdef USE_VERTEX_AREA_WEIGHTING
    std::map<unsigned char, float> neighborhoodHistogram;
#else
    std::map<unsigned char, unsigned> neighborhoodHistogram;
#endif

    unsigned char binnedValue;

    float c, p, score = 0.0f;

    const float range = max - min;

#ifdef USE_VERTEX_AREA_WEIGHTING
    float summedArea = 0.0f;

    for (std::vector<lce::TriMesh::VertexHandle>::const_iterator nIt = neighbors.begin(), nEnd = neighbors.end(); nIt != nEnd; ++nIt)
    {
        summedArea += mesh.property(vertexAreasProperty, *nIt);
    }
#endif

    for (std::vector<lce::TriMesh::VertexHandle>::const_iterator nIt = neighbors.begin(), nEnd = neighbors.end(); nIt != nEnd; ++nIt)
    {
        c = std::max(std::min(mesh.property(curvatureMeasure, *nIt), max), min);

        binnedValue = static_cast<unsigned char>(255.0f * (c - min) / range);

#ifdef USE_VERTEX_AREA_WEIGHTING
        neighborhoodHistogram[binnedValue] += mesh.property(vertexAreasProperty, *nIt);
#else
        ++neighborhoodHistogram[binnedValue];
#endif
    }

#ifdef USE_VERTEX_AREA_WEIGHTING
    for (std::map<unsigned char, float>::iterator mIt = neighborhoodHistogram.begin(), mEnd = neighborhoodHistogram.end(); mIt != mEnd; ++mIt)
#else
    for (std::map<unsigned char, unsigned>::iterator mIt = neighborhoodHistogram.begin(), mEnd = neighborhoodHistogram.end(); mIt != mEnd; ++mIt)
#endif
    {
#ifdef USE_VERTEX_AREA_WEIGHTING
        p      =  mIt->second / summedArea;
#else
        p      = static_cast<float>(mIt->second) / neighbors.size();
#endif
        score += p * -log(p)/log(2.0f);
    }

    return score;
}

//--------------------------------------------------------------------------------------------------------------

void simpleSmooth(lce::TriMesh & mesh, OpenMesh::VPropHandleT<float> property)
{
    std::map<lce::TriMesh::VertexHandle, float> averageVal;

    lce::TriMesh::VertexIter vIt, vEnd;
    lce::TriMesh::VertexVertexIter vvIt;

    unsigned count;
    float average;

    for (vIt = mesh.vertices_begin(), vEnd = mesh.vertices_end(); vIt != vEnd; ++vIt)
    {
        average = mesh.property(property, *vIt);
        count   = 1;

        for (vvIt = mesh.cvv_iter(*vIt); vvIt.is_valid(); ++vvIt, ++count)
        {
            average += mesh.property(property, *vvIt);
        }

        average /= count;

        averageVal[*vIt] = average;
    }

    for (vIt = mesh.vertices_begin(), vEnd = mesh.vertices_end(); vIt != vEnd; ++vIt)
    {
        mesh.property(property, *vIt) = averageVal[*vIt];
    }
}

//--------------------------------------------------------------------------------------------------------------

namespace  lce
{

//--------------------------------------------------------------------------------------------------------------

bool estimateSaliency(TriMesh & mesh)
{
    std::clock_t overallBegin = std::clock();
    std::clock_t timerBegin;
    std::clock_t ticksNeighborhoodSearch  = 0,
                 ticksSaliencyComputation = 0;


    //compute per-vertex curvature first

    OpenMesh::VPropHandleT<float> gaussianCurvatureProperty;
    OpenMesh::VPropHandleT<float> meanCurvatureProperty;

    computeCurvature(mesh, gaussianCurvatureProperty, meanCurvatureProperty);

    //store result in vertex colors
    mesh.request_vertex_colors();


    TriMesh::VertexIter vIt, vEnd;

    std::vector<TriMesh::VertexHandle> neighbors;

    float s;


    //build curvature histogram

    Histogram hCurvature;

    for (vIt = mesh.vertices_begin(), vEnd = mesh.vertices_end(); vIt != vEnd; ++vIt)
    {
        hCurvature.add(mesh.property(meanCurvatureProperty, *vIt));
    }


    //compute or read saliency

    OpenMesh::VPropHandleT<float> vertexAreasProperty = computeVertexAreas(mesh);

    //set radius, which is the most important parameter (regarding result and also performance)
    const float SurfaceArea = surfaceArea(mesh);
    const float MaxRadius   = 0.08f * sqrt(SurfaceArea);

#ifdef USE_VERTEX_AREA_WEIGHTING
    FloatHistogram hSaliency;
#else
    Histogram hSaliency;
#endif
    
    //we need edge lengths to compute the geodesic neighborhood
    OpenMesh::EPropHandleT<float> edgeLengthProperty = computeEdgeLengths(mesh);

    OpenMesh::VPropHandleT<float> saliencyProperty;

    //compute saliency, if no quality property is already given - otherwise, use given quality
    //this can be used to compare different methods for saliency estimation, using the same color mapping
    if (mesh.get_property_handle(saliencyProperty, "quality"))
    {
        std::cout << "Found quality property - not computing saliency, but applying same color mapping, using given quality instead." << std::endl;

        for (vIt = mesh.vertices_begin(), vEnd = mesh.vertices_end(); vIt != vEnd; ++vIt)
        {
            s = mesh.property(saliencyProperty, *vIt);

#ifdef USE_VERTEX_AREA_WEIGHTING
            hSaliency.add(s, 1.0f);
#else
            hSaliency.add(s);
#endif
        }
    }    
    //no saliency property already given -> compute saliency
    else
    {
        mesh.add_property(saliencyProperty);

        OpenMesh::VPropHandleT< std::vector<float> > multiscaleSaliencyProperty;
        mesh.add_property(multiscaleSaliencyProperty);

        //compute saliency at multiple levels
        unsigned currentLevel = 0;

        while (currentLevel++ < NUM_SALIENCY_LEVELS)
        {
            std::cout << "Saliency computation (lv " << currentLevel << "/" << NUM_SALIENCY_LEVELS << "): 0%";
            std::cout.flush();

            float percentage = 0.0f;
            unsigned n       = 0;

            const float currentRadius = MaxRadius / pow(2.0f, static_cast<float>(currentLevel - 1));

            for (vIt = mesh.vertices_begin(), vEnd = mesh.vertices_end(); vIt != vEnd; ++vIt, ++n)
            {
                timerBegin = std::clock();

                getGeodesicNeighborhood(mesh, *vIt, currentRadius, edgeLengthProperty, neighbors);

                ticksNeighborhoodSearch += std::clock() - timerBegin;


                timerBegin = std::clock();

                s = computeSaliency(mesh, *vIt, neighbors, vertexAreasProperty,
                                    meanCurvatureProperty, hCurvature.getMin(), hCurvature.getMax());

                ticksSaliencyComputation += std::clock() - timerBegin;


                mesh.property(multiscaleSaliencyProperty, *vIt).push_back(s);


                neighbors.clear();

                if (floor(100.0f * (static_cast<float>(n) / mesh.n_vertices())) > percentage)
                {
                    percentage = floor(100.0f * (static_cast<float>(n) / mesh.n_vertices()));

                    std::cout << "\r";
                    std::cout.flush();
                    std::cout << "Saliency computation (lv " << currentLevel << "/" << NUM_SALIENCY_LEVELS << "): " << static_cast<unsigned>(percentage) << "%";
                }
            }

            std::cout << "\r";
            std::cout.flush();
            std::cout << "Saliency computation (lv " << currentLevel << "/" << NUM_SALIENCY_LEVELS << "): 100%" << std::endl;
        }

        //merge saliency values to a single score / property
        for (vIt = mesh.vertices_begin(), vEnd = mesh.vertices_end(); vIt != vEnd; ++vIt)
        {
            s = 0.0f;

            for (unsigned i = 0; i < NUM_SALIENCY_LEVELS; ++i)
            {
                const std::vector<float> & saliencyValues = mesh.property(multiscaleSaliencyProperty, *vIt);

                assert(saliencyValues.size() == NUM_SALIENCY_LEVELS);

                s += saliencyValues[i];
            }

            s /= NUM_SALIENCY_LEVELS;

            mesh.property(saliencyProperty, *vIt) = s;
        }


        //smooth values to remove noise
        {
            unsigned smoothingSteps = 4;

            while (smoothingSteps--)
            {
                simpleSmooth(mesh, saliencyProperty);
            }
        }


        //normalize
        float sMin =  FLT_MAX;
        float sMax = -FLT_MAX;
        for (vIt = mesh.vertices_begin(), vEnd = mesh.vertices_end(); vIt != vEnd; ++vIt)
        {
            s = mesh.property(saliencyProperty, *vIt);

            sMin = std::min(s, sMin);
            sMax = std::max(s, sMax);
        }
        float sRange = sMax - sMin;
        for (vIt = mesh.vertices_begin(), vEnd = mesh.vertices_end(); vIt != vEnd; ++vIt)
        {
            s = mesh.property(saliencyProperty, *vIt);

            mesh.property(saliencyProperty, *vIt) = (s - sMin) / sRange;
        }


        //build histogram
        for (vIt = mesh.vertices_begin(), vEnd = mesh.vertices_end(); vIt != vEnd; ++vIt)
        {
            s = mesh.property(saliencyProperty, *vIt);

#ifdef USE_VERTEX_AREA_WEIGHTING
            hSaliency.add(s, mesh.property(vertexAreasProperty, *vIt));
#else
            hSaliency.add(s);
#endif
        }

        std::cout << "Saliency computation finished." << std::endl;
        std::cout << "(took "                  << double(std::clock() - overallBegin) / CLOCKS_PER_SEC << " seconds)" << std::endl;
        std::cout << "(neighborhood search:  " << double(ticksNeighborhoodSearch)     / CLOCKS_PER_SEC << " seconds)" << std::endl;
        std::cout << "(saliency computation: " << double(ticksSaliencyComputation)    / CLOCKS_PER_SEC << " seconds)" << std::endl;
    }


    //cleanup helper structures

    mesh.remove_property(edgeLengthProperty);


    //colorize mesh vertices

    //float mappingMin = hSaliency.getPercentile(0.01f);
    //float mappingMax = hSaliency.getPercentile(0.99f);

    float mappingMin = hSaliency.getMin();
    float mappingMax = hSaliency.getMax();


    const float colVScale = 255.0f;

    OpenMesh::Vec3f color;
    TriMesh::Color  colorCharV;

    for (vIt = mesh.vertices_begin(), vEnd = mesh.vertices_end(); vIt != vEnd; ++vIt)
    {
        s = std::min(std::max(mesh.property(saliencyProperty, *vIt), mappingMin), mappingMax);

        mapScalarToColor(color, (s - mappingMin) / (mappingMax - mappingMin));

        colorCharV[0] = color[0] * colVScale;
        colorCharV[1] = color[1] * colVScale;
        colorCharV[2] = color[2] * colVScale;

        mesh.set_color(*vIt, colorCharV);
    }

    OpenMesh::VPropHandleT<float> qualityProperty;
    if (!mesh.get_property_handle(qualityProperty, "quality"))
    {
        mesh.add_property(qualityProperty, "quality");

        for (vIt = mesh.vertices_begin(), vEnd = mesh.vertices_end(); vIt != vEnd; ++vIt)
        {
            mesh.property(qualityProperty, *vIt) = mesh.property(saliencyProperty, *vIt);
        }
    }
    //make sure the property is written to the file
    mesh.property(qualityProperty).set_persistent(true);

    return true;
}

//------------------------------------------------------------------------------------------------------------------

} //namespace lce
